/*

Hmm.. I wonder what happens if I take a bunch of broken cobbled together routines for hexagonal grids and add a drawtool interface. Look... a fun game to play with your loved ones.

Link this against libm
At the moment, I cba to do anything clever with keypresses. So, the main loop runs through getc() lots. I use quickbutton to generate the "x\n" or "z\n" or "f\n" for left, right and fire.

quickbutton "z" "x" "f" | ./hex |drawtool -g 384x550+0+0

*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#define LEVELPATH "./levels/"

#define XSIZE 8 
#define YSIZE 12

#define COLLAPSETIME 8 // how many shots before ceiling falls

// 8:7 is good. 24:21 is a good size
#define RADIUS 8
#define YSTEP 7

#define GRIDWIDTH (RADIUS*XSIZE*2)
#define GRIDHEIGHT ((YSTEP*YSIZE*2+YSTEP/2)-4)

#define EMPTY 0
#define RED 1
#define GREEN 2
#define BLUE 3
#define GREY 4

#define PLAYERX (GRIDWIDTH/2)
#define PLAYERY (GRIDHEIGHT+YSTEP*2)
#define PLAYERLEN (YSTEP*4) // looks about right

struct grid
	{
	char cell[XSIZE][YSIZE];
	int ceiling; // 0 for normal, YSIZE-1 for crushed
	};
typedef struct grid grid;

struct player
	{
	int angle;
	int currentColour;
	int shots;
	};
typedef struct player player;

void error(char *s);
grid *newGrid(void);
void destroyGrid(grid *g);
void initGrid(grid *g);
void printGrid(grid *g);
player *newPlayer(void);
void destroyPlayer(player *p);
void initPlayer(player *p);
void printPlayer(player *p);
void loadLevel(grid *g,int level);
void hexToSquare(int hx,int hy,int *sx,int *sy);
void shootPlayer(player *p,grid *g);
int touching(grid *g,int x1,int y1,int x2,int y2);
void checkDrop(grid *g,player *p,int startx,int starty);
void checkMultiple(grid *g,player *p,int oldtx,int oldty,int *xs,int *ys, int *count,int sameColour);
int countTouching(grid *g,int startx,int starty);
void shiftDown(grid *g);
void checkForEnd(player *p, grid *g);
void highlightTouching(grid *g,int startx,int starty);
void listTouching(grid *g,int tx,int ty);
void compactList(int *xs,int *ys,int *origCount);


void loadLevel(grid *g,int level)
{
char filename[1024];
FILE *fp;
int x,y;
int c;
int colour;

sprintf(filename,"%s/Level%d",LEVELPATH,level);

fp = fopen(filename,"r");
if (fp==NULL) error("Level file missing");

x=0;
y=0;
while((c=fgetc(fp))!=EOF)
	{
	if (isalpha(c))
		{
		colour=-1;
		switch(c)
			{
			case 'R':colour=RED;break;
			case 'G':colour=GREEN;break;
			case 'B':colour=BLUE;break;
			}
		if (colour==-1)
			{
			error("Level file full of crap");
			}
		g->cell[x][y] = colour;
		x++;
		if (x==XSIZE-(y%2))
			{
			x=0;
			y++;
			}
		}
	}
fclose(fp);
}

void error(char *s)
{
fprintf(stderr,"\n%s\n\n",s);
exit(-1);
}

grid *newGrid(void)
{
grid *g;
if ( (g=(grid *)malloc(sizeof(grid))) == NULL ) error("Out of memory");
return g;
}

player *newPlayer(void)
{
player *p;
if ( (p=(player *)malloc(sizeof(player))) == NULL ) error("Out of memory");
return p;
}

void destroyGrid(grid *g)
{
free(g);
}

void destroyPlayer(player *p)
{
free(p);
}

void initGrid(grid *g)
{
int x,y;

g->ceiling = 0;

for (x=0;x<XSIZE;x++){
for (y=0;y<YSIZE;y++){
g->cell[x][y] = EMPTY;
}}

fprintf(stderr,"%d,%d\n",GRIDWIDTH,GRIDHEIGHT);
}

void initPlayer(player *p)
{
p->angle=180; // straight up
p->currentColour=EMPTY;
p->shots=0;
}

void printGrid(grid *g)
{
int x,y,xstep,c;
int ri,gi,bi;
int n;

for(x=0;x<XSIZE;x++){
for(y=0;y<YSIZE;y++){
if (g->cell[x][y] < 0 || g->cell[x][y]>GREY) error("SHIT");
}}

printf("3,%d,%d,%d,%d,65535,65535,65535\n",0,0,GRIDWIDTH,PLAYERY);
printf("2,%d,%d,%d,%d,0,0,0\n",0,0,GRIDWIDTH,GRIDHEIGHT);
fflush(stdout);

printf("1,%d,%d,%d,%d,0,0,0\n",0,g->ceiling*YSTEP*2,GRIDWIDTH,g->ceiling*YSTEP*2);

for (y=0;y<YSIZE;y++){
if (g->ceiling%2) n = !(y%2); else n = (y%2);
//for (x=0;x<XSIZE-(n);x++) // if odd row, one less
for (x=0;x<XSIZE;x++)
	{
	xstep = (n)*RADIUS;
	switch(g->cell[x][y])
		{
		case EMPTY:
//		printf("%d,%d,%d,%d,%d,%d,%d\n",5,(x*RADIUS*2)+xstep+RADIUS,(y*YSTEP*2)+RADIUS,RADIUS,65535,65535,65535);
//		fflush(stdout);
		c=-1;ri=0;gi=0;bi=0;
		break;
		case RED:
		c=1;ri=65535;gi=0;bi=0;
		break;
		case GREEN:
		c=1;ri=0;gi=65535;bi=0;
		break;
		case BLUE:
		c=1;ri=0;gi=0;bi=65535;
		break;
		case GREY:
		c=1;ri=32768;gi=32768;bi=32768;
		break;
		default:
		c=0;ri=0;gi=0;bi=0;
		break;
		}
	if (c!=-1)
		{
		printf("%d,%d,%d,%d,%d,%d,%d\n",4+c,(x*RADIUS*2)+xstep+RADIUS,(y*YSTEP*2)+RADIUS,RADIUS,ri,gi,bi);
		fflush(stdout);
		}
	}}

fflush(stdout);
}

void printPlayer(player *p)
{
	int x,y; // endpoint of barrel

	x = PLAYERX + sin(p->angle*3.14159/180.0)*PLAYERLEN;
	y = PLAYERY + cos(p->angle*3.14159/180.0)*PLAYERLEN;

	printf("1,%d,%d,%d,%d,%d,%d,%d\n",PLAYERX,PLAYERY,x,y,0,0,0);
        printf("5,%d,%d,%d,%d,%d,%d\n",PLAYERX,PLAYERY,RADIUS,(p->currentColour==RED)*65535,(p->currentColour==GREEN)*65535,(p->currentColour==BLUE)*65535);
	fflush(stdout);
}

void shootPlayer(player *p,grid *g)
{
float dx;
float dy;
float x,y;
int stop=0;
int tx=-1,ty=-1;
int count=0;
int xs[XSIZE*XSIZE];
int ys[YSIZE*YSIZE];
int i;

x = PLAYERX;
y = PLAYERY;

dx = sin(p->angle*3.14159/180.0)*(YSTEP/2); // XSTEP/2 is the speed
dy = cos(p->angle*3.14159/180.0)*(YSTEP/2); // XSTEP/2 is the speed

while(!stop)
	{
	usleep(300);
	printf("0,%d,%d,%d,%d,%d\n",(int)x,(int)y,(p->currentColour==RED)*65535,(p->currentColour==GREEN)*65535,(p->currentColour==BLUE)*65535);
	fflush(stdout);

	x += dx;
	y += dy;

	hexToSquare((int)x,(int)y,&tx,&ty);

	if (ty<=g->ceiling) //hit the roof
		{
		stop=1;
		}
	else
//	if (g->ceiling%2) n = !(ty%2); else n = (ty%2);
	if (countTouching(g,tx,ty)) // touching something
		{
//		highlightTouching(g,tx,ty);
//fprintf(stderr,"Touching: %d",countTouching(g,tx,ty));
//listTouching(g,tx,ty);
		printGrid(g);
		if (ty>=YSIZE-1)
			{
			fprintf(stderr,"\nGame Over, hit bottom");
			exit(-1);
			}
		stop=1;
		}
	if ((x<=0 || x>=GRIDWIDTH-RADIUS)) dx=-dx; 
} // end while

//if (oldtx==-1 || oldty==-1) error("Oh dear");

g->cell[tx][ty] = p->currentColour; // collision
//fprintf(stderr,"\nConnect: %d,%d",oldtx,oldty);

xs[0] = tx;
ys[0] = ty;
count++;

checkMultiple(g,p,tx,ty,xs,ys,&count,1);

if (count>3)
	{
	compactList(xs,ys,&count);
	for (i=0;i<count;i++)
		{
//		fprintf(stderr,"\nRemove:%d, %d",xs[i],ys[i]);
		g->cell[xs[i]][ys[i]] = EMPTY;
		checkDrop(g,p,xs[i],ys[i]);
		}
	}
}

void listTouching(grid *g,int tx,int ty)
{
int x,y;
int n;

for(x=tx-1;x<=tx+1;x++){
for(y=ty-1;y<=ty+1;y++){
if (g->cell[x][y]!=EMPTY)
	{
	n=touching(g,x,y,tx,ty);
//	fprintf(stderr,"\nTouching Number: %d",n);
	}
}}
}

void checkForEnd(player *p, grid *g)
{
int x,y,end=1;

for(x=0;x<XSIZE;x++){
for(y=0;y<YSIZE;y++){
if (g->cell[x][y]!=EMPTY) end=0;
}}

if (end)
	{
	fprintf(stderr,"\nWell done, you finished in %d shots",p->shots);
	printGrid(g);
	exit(0);
	}
}

int countTouching(grid *g,int startx,int starty)
{
int x,y,i=0;
int n;
int j;

for (x=startx-1;x<=startx+1;x++){
for (y=starty-1;y<=starty+1;y++){
	if (g->ceiling%2) n = !(y%2); else n = (y%2);
	if (x>=0 && y>=0 && x<XSIZE-n && y<YSIZE)
		{
		if (g->cell[x][y]!=EMPTY && (j=touching(g,startx,starty,x,y)>0))
			{
			//fprintf(stderr,"\ncounted: %d",j);
			i++;
			}
		}
}}
if (i>6) error("Too much touching");
return i;
}

void highlightTouching(grid *g,int startx,int starty)
{
int x,y;
grid *t;

fprintf(stderr,"\nHighlight touching: %d,%d",startx,starty);

t = newGrid();
initGrid(t);

for (x=0;x<XSIZE;x++){
for (y=0;y<YSIZE;y++){ // don't run over
t->cell[x][y] = g->cell[x][y];
t->ceiling = g->ceiling;
}}

for (x=startx-10;x<=startx+10;x++){
for (y=starty-10;y<=starty+10;y++){
if (touching(g,startx,starty,x,y)>0) t->cell[x][y]=GREY;
}}

printGrid(t);
//usleep(65535);
//sleep(2);

destroyGrid(t);
}

void checkDrop(grid *g,player *p,int startx,int starty)
{
int i,j;
int x,y;
int xs[(XSIZE*YSIZE)*(XSIZE*YSIZE)];
int ys[(XSIZE*YSIZE)*(XSIZE*YSIZE)];
int count;
int stuck;
int n;

for (y=0;y<YSIZE;y++){ // one less on odd rows
if (g->ceiling%2) n = !(y%2); else n = (y%2);
for (x=0;x<XSIZE-n;x++)
	{
	if (g->cell[x][y]!=EMPTY)
	{
	count = 0;
	checkMultiple(g,p,x,y,xs,ys,&count,0);
	if (count)
		{
		stuck = 0;
		for (i=0;i<count;i++) if (ys[i] == g->ceiling) stuck=1;
		if (!stuck)
			{
			for (j=0;j<count;j++)
				{
				if (ys[j] == g->ceiling) error("ug1");
//				fprintf(stderr,"\nRemove %d,%d",xs[j],ys[j]);
				g->cell[xs[j]][ys[j]] = EMPTY;
				}
			}
		}
	}
	if (g->cell[x][y]!=EMPTY && countTouching(g,x,y)==0 && y>g->ceiling)
		{
//		fprintf(stderr,"\nRemove: %d,%d",x,y);
		g->cell[x][y]=EMPTY; // clean up strays
		checkDrop(g,p,startx,starty); // look for more
		}
}
}
}

void checkMultiple(grid *g,player *p,int oldtx,int oldty,int *xs,int *ys, int *count,int sameColour)
{
int x,y;
int i,j;
int f;
int n;

for (x=oldtx-1;x<=oldtx+1;x++){
for (y=oldty-1;y<=oldty+1;y++){
//fprintf(stderr,"\n%d,%d",x,y);
if (sameColour) // must be the same colour
        { 
        f = (g->cell[x][y]==g->cell[oldtx][oldty]);     
        } 
else f=1;

if (g->ceiling%2) n = !(y%2); else n = (y%2);
if ((x<XSIZE-n && y<YSIZE && x>=0 && y>=0) && f && touching(g,oldtx,oldty,x,y)>0 && g->cell[x][y]!=EMPTY)
	{
	xs[*count] = x; // store current collision
	ys[*count] = y;
	(*count)++;
//	fprintf(stderr,"\nTouching: %d,%d",x,y);

	j=0; // check if x,y is already in list to be removed
       	for (i=0;i<(*count)-1;i++)
               	{
 //              	if ((xs[i]==x && ys[i]==y)||(xs[i]==oldtx && ys[i]==oldty))
                if ((xs[i]==x && ys[i]==y))
                     	{
//                       	fprintf(stderr,"\nAlready in list");
                       	j=1;
                       	}
               	}       
	if (!j)
		{
		compactList(xs,ys,count);
		checkMultiple(g,p,x,y,xs,ys,count,sameColour);
		}
	}
	}}
}

void compactList(int *xs,int *ys,int *origCount)
{
// this is so bad it makes me cry. This function saves on a bit of memory, but takes a billion cycles to complete. Need to find a nice one from Knuth or something for removing duplicates from lists.

int newxs[500];
int newys[500];
int newcount;
int count;
int i,j;
int found;

count=*origCount;
if (count>=500) error("Bugger me backwards with a fishfork");

newcount=0;
for(i=0;i<count;i++)
	{
	found = 0;
	for (j=0;j<i;j++)
		{
		if (newxs[j] == xs[i] && newys[j] == ys[i])
			{
			found = 1; // break
			}
		}
	if (!found)
		{
		newxs[newcount] = xs[i];
		newys[newcount] = ys[i];
		newcount++;
		}
	}
//fprintf(stderr,"\nnewcount: %d\ncount: %d",newcount,count);

for(i=0;i<newcount;i++) // copy back
	{
	xs[i] = newxs[i];
	ys[i] = newys[i];
	}
*origCount = newcount;
}

int touching(grid *g,int x1,int y1,int x2,int y2)
{
int n;

// this works!!!!!!! OMIGOD

if (!(g->ceiling%2))
{
if (y1%2) n=1;
else n=-1;
}
else
{
if (y1%2) n=-1;
else n=1;
}

//fprintf(stderr,"\nx1,y1 = %d,%d\nx2,y2 = %d,%d",x1,y1,x2,y2);
if (x1<0 || y1<0 || x2<0 || y2<0) return -2;
if (x1>XSIZE-n || x2>XSIZE-n || y1>YSIZE-1 || y2>YSIZE-1) return -1;
if (x1==x2 && y1==y2) return -3;

if (x1+n==x2 && y1-1==y2) return 1;
if (x1==x2 && y1-1==y2) return 2;
if (x1-1==x2 && y1==y2) return 3;
if (x1+1==x2 && y1==y2) return 4;
if (x1+n==x2 && y1+1==y2) return 5;
if (x1==x2 && y1+1==y2) return 6;
return 0;
}

void hexToSquare(int hx,int hy,int *sx,int *sy)
{
int shift=0;

*sx = hx/(RADIUS*2);
if ((*sy)%2) shift=RADIUS;
*sy = (hy-shift)/(YSTEP*2);

//if (*sx >= XSIZE || *sy >= YSIZE) fprintf(stderr,"\nFreak Out!: sx=%d,sy=%d,hx=%d,hy=%d",*sx,*sy,hx,hy);

if (*sy >= YSIZE) *sy=YSIZE-1;
}

void shiftDown(grid *g)
{
int x,y;
grid *t;

t = newGrid();
initGrid(t);

g->ceiling++;

for (x=0;x<XSIZE;x++){
for (y=0;y<YSIZE-1;y++){ // don't run over
t->cell[x][y+1] = g->cell[x][y];
t->ceiling = g->ceiling;
}}

for (x=0;x<XSIZE;x++){
for (y=0;y<YSIZE;y++){
g->cell[x][y] = t->cell[x][y];
g->ceiling = t->ceiling;
}}

destroyGrid(t);
}


void main(void)
{
grid *g;
player *p;
int c;

g = newGrid();
initGrid(g);
loadLevel(g,1);

p = newPlayer();
initPlayer(p);

p->currentColour=(rand()%BLUE)+1;

printGrid(g);
printPlayer(p);

srand(42);

//shiftDown(g);

//highlightTouching(g,3,3);

while(1)
{
c = fgetc(stdin);

if (c==120)
	{
	p->angle-=5;
	if (p->angle<95) p->angle=95;
	printPlayer(p);
	}

if (c==122)
        {
        p->angle+=5;
        if (p->angle>265) p->angle=265;
	printPlayer(p);
        }

if (c=='f')
	{
	shootPlayer(p,g);
	p->shots++;
	p->currentColour=(rand()%BLUE)+1;
	if (p->shots%COLLAPSETIME == 0) shiftDown(g);
	checkForEnd(p,g);
	}
printGrid(g);
printPlayer(p);
}
}
